/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: WinLib.h                                                                                       |
|                Author: Michael Krmer                                                                                 |
|                        Michael@dk8pp.de                                                                               |
|                  Date: 03-Dec-2014                                                                                    |
|              Language: Visual C Express 2013                                                                          |
|               Version: 2.1                                                                                            |
|        Program status: Release                                                                                        |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Library with some often used functions for Windows programming                                              |
|                                                                                                                       |
|  History: V1.0, 30-Nov-2007: First public release                                                                     |
|           V1.1, 29-May-2011: Adaptation to stdint.h and stdbool.h, remove mytype.h                                    |
|           V2.1, 08-Dec-2021: Resize and Move improved                                                                 |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifndef WINLIB
#include <stdint.h>
#include <stdbool.h>

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define WINLIB 1
#define MY_REG_KEY_ROOT _T("Software\\DK8PP Freeware\\") /* Root for registry keys under HKEY_CURRENT_USER */
#define MAX_REG_KEY_LENGHT 128  /* max length of registry key (MY_REG_KEY_ROOT + sub keynames) */
#define PARM_SIMPLE 1000        /* simple parameter, no argument (eg. -a, -rc etc.) */
#define PARM_INT 1001           /* integer parameter (eg. -q=-25) */
#define PARM_STRING 1002        /* string parameter (eg. -f=image.jpg, -f="c:\Program Files\image.jpg") */
#define PARM_POS 1003           /* positional parameter (eg. image.jpg, "c:\Program Files\image.jpg") */
#define MAXPARMLENGTH MAX_PATH  /* maximum length of a command line parameter, could be a path name */
#define INI_NO 0                /* no initialization data */
#define INI_REGREAD   1         /* read registry */
#define INI_REGWRITE  2         /* write registry */
#define INI_FILEREAD  4         /* read initialization file */
#define INI_FILEWRITE 8         /* write initialization file */
#define REGFLAG_NONE 0          /* no flag */
#define REGFLAG_WRITEBACK 1     /* this registry entry is updated when the program executes and written back when it terminates */
#define REGFLAG_ARRAY  0x10000  /* this registry entry is an array of (u)int32_t or (u)int64_t */
#define REGFLAG_HEX    0x20000  /* read and write as a hex number */
#define REGFLAG_SIGNED 0x40000  /* read and write as a int32_t or int64_t */
#define SYSP _T("$")            /* System command line parameters start with this character */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/
typedef struct {
        uint32_t control ;      /* ID */
        HWND     *hwnd ;        /* pointer to dialog box handle */
        int xoffs ;             /* additional horizontal offset */
        int yoffs ;             /* additional vertical offset */
        int left ;              /* left position of control */
        int top ;               /* top position of control */
        int right ;             /* right position of control */
        int bottom ;            /* bottom position of control */
        int movespeed_x ;       /* move speed in x direction (0..1000) */
        int movespeed_y ;       /* move speed in y direction (0..1000) */
        int resizespeed_x ;     /* resize speed in x direction (0..1000) */
        int resizespeed_y ;     /* resize speed in y direction (0..1000) */
                } COORDS ;

typedef struct { /* new, works with registry or ini-file */
        LPTSTR sectionname ;    /* points to the name of the section */
        LPTSTR valuename ;      /* points to the name of the value */
        uint32_t valuetype ;    /* type of the value */
        uint32_t flags ;        /* optional flags, see REGFLAG_ */
        void *value ;           /* points to the value */
        size_t valuesize ;      /* buffer size of the value for type REG_MULTI_SZ in number of characters including terminating 0.
                                   Number of words for REG_DWORD/REG_QWORD & REGFLAG_ARRAY (should be 1 for REG_DWORD/REG_QWORD) */
                } INIVALUE ;

typedef struct { /* new, works with registry or ini-file */
        HKEY hkey ;             /* handle of open key */
        LPTSTR inikeyname ;     /* pointer to name of initialization key */
        LPTSTR inifilename ;    /* pointer to name of ini-file */
        INIVALUE *inivalues ;   /* pointer to the array of INIVALUEs */
        uint32_t initype ;      /* type of initialization (i.e. registry or ini-file) */
        size_t size ;           /* number of initialization values */
                } MYINI ;

typedef struct { /* command line options */
        LPTSTR parmname ;       /* name of parameter without leading "-" or "/". System parameters start with SYSP character. */
        uint32_t parmtype ;     /* type of this parameter */
        void *dest ;            /* pointer to destination for a string parameter */
        int32_t value ;         /* value of this parameter or maximum stringlength in case of a string */
        bool found ;            /* parameter was found */
        LPTSTR help ;           /* help text for this parameter */
                } CLO ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void display_info(LPTSTR msg) ;
void ShowError(int dialog) ;
void ShowSystemError(int errnr, int dialog) ;
void resize_controls(COORDS *coordinates, size_t size) ;
void get_coords(COORDS *coordinates, size_t size) ;
int32_t compare_hup(const void *op1, const void *op2) ;
int32_t compare_hdown(const void *op1, const void *op2) ;
int32_t compare_vup(const void *op1, const void *op2) ;
int32_t compare_vdown(const void *op1, const void *op2) ;
INT_PTR WinLibMessage(unsigned msg, WPARAM wparam, LPARAM lparam) ;
void WinLibInit(void) ;
void SaveCoordinates(void) ;
bool get_registry(MYINI *myreg) ;
void put_registry(MYINI *myreg) ;
void clear_registry(MYINI *myreg) ;
void PrintProgramInfo(uint32_t dlgitem, LPTSTR programname, LPTSTR date, LPTSTR time) ;
LPTSTR getparms(LPTSTR cmdline) ;
void set_process_prio(int32_t prio) ;
void set_thread_prio(HANDLE hthread, int32_t prio) ;
void *my_malloc(size_t size, LPTSTR file, int line, int aux) ;
void my_free(void *ptr) ;
void PrintInfo(LPTSTR info) ;

#endif

