﻿/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: DrawTimingDiagram                                                                              |
|                Author: Michael Krämer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 04.12.2021                                                                                     |
|              Language: Microsoft Visual Studio 2019                                                                   |
|               Version: 1.01                                                                                           |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Generate text output timing diagrams in UTF-8                                                               |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+

Preprocessor-Macros
===================

NDEBUG:     NDEBUG is a standard macro with the semantic "Not Debug" for the C and C++ standards.
            If NDEBUG is undefined, debug code is enabled. NDEBUG is used by the assert macro.

_DEBUG:     Visual Studio automatically defines _DEBUG /MTd or /MDd option.

*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <time.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <tchar.h>
#include <stdbool.h>
#include <process.h>
#include <assert.h>
#include "resource.h"
#include "WinLib.h"

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define PARMSIZE 64             /* maximum size of a string or positional command line parameter */
#define STRINGSIZE 1024         /* default size for strings */
#define MAX_SIGNALNAME 64       /* maximum length of a signal name */
#define COMMANDCHAR _T('$')     /* this character denotes a command */

#define MY_HKEY HKEY_CURRENT_USER /* handle of registry key */
#define MY_INI_SECT _T("DrawTimingDiagram")    /* Section name or registry key used by this program */
#define MY_REG_KEY MY_REG_KEY_ROOT MY_INI_SECT /* Registry key used by this program */
#define PROGRAMNAME _T("DrawTimingDiagram")
#define PURPOSE _T("This program generates text mode timing diagrams.\r\r\n\r\r\n")
#ifdef _DEBUG
#define DEBUG_T _T("-Debug")
#else
#define DEBUG_T _T("")
#endif
#ifdef _WIN64
#define INTSIZE_T _T("64-bit")
#else
#define INTSIZE_T _T("32-bit")
#endif
#define PROGRAMVERSION 101      /* will be displayed as n.nn */
#define DLG_STATUS (WM_APP+0)   /* Status */
#define STATUSMESSAGE 0         /* Part of status window for status message */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       external variables                                                                                              |
+-----------------------------------------------------------------------------------------------------------------------+
*/
extern int32_t snap_x, snap_y ; /* horizontal and vertical snap distances */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
HINSTANCE       MyInstance ;
HWND            MainWindow ;
HWND            StatusWindow ;
static _TCHAR strng1[STRINGSIZE] ;
static _TCHAR strng2[STRINGSIZE] ;
static _TCHAR ini_filename[MAX_PATH+1] ;

_TCHAR programname[] = PROGRAMNAME ;
int programversion = PROGRAMVERSION ;

/*
 Registry definitions
*/
static int32_t registry_x = 0 ;                         /* left edge of dialog */
static int32_t registry_y = 0 ;                         /* top edge of dialog */
static int32_t registry_width = 0 ;                     /* width of dialog */
static int32_t registry_height = 0 ;                    /* height of dialog */
static int32_t registry_version = PROGRAMVERSION ;      /* version of program which wrote the registry */
static _TCHAR td_font[MAX_PATH+1] = _T("Courier New") ; /* Timing Diagram: font for input and output windows */
static uint32_t td_fontsize = 16 ;                      /* Timing Diagram: font size */
static uint32_t td_fontweight = 700 ;                   /* Timing Diagram: font weight (0 .. 1000, 400 = normal, 700 = bold) */
static uint32_t td_textlimit = 32*1024 ;                /* Timing Diagram: maximum number of characters in input/output edit controls */

//                             ,-------------------------------------------------------------------------------------------------- section name
//                             |                    ,----------------------------------------------------------------------------- value name
//                             |                    |                       ,----------------------------------------------------- value type
//                             |                    |                       |          ,------------------------------------------ optional flags, see REGFLAG_
//                             |                    |                       |          |                  ,----------------------- points to the value
//                             |                    |                       |          |                  |                  ,---- buffer size of the value for type REG_MULTI_SZ in number of characters including terminating 0.
//                             |                    |                       |          |                  |                  |     Number of words for REG_DWORD&REGFLAG_ARRAY (should be 1 for REG_DWORD)
//                             |                    |                       |          |                  |                  |
static INIVALUE inivalue[] = {{_T("Window"),        _T("PosX"),             REG_DWORD, REGFLAG_WRITEBACK, &registry_x,       1},
                              {_T("Window"),        _T("PosY"),             REG_DWORD, REGFLAG_WRITEBACK, &registry_y,       1},
                              {_T("Window"),        _T("Width"),            REG_DWORD, REGFLAG_WRITEBACK, &registry_width,   1},
                              {_T("Window"),        _T("Height"),           REG_DWORD, REGFLAG_WRITEBACK, &registry_height,  1},
                              {_T("Program"),       _T("Version"),          REG_DWORD, REGFLAG_NONE,      &registry_version, 1},
                              {_T("TimingDiagram"), _T("Font"),             REG_SZ,    REGFLAG_WRITEBACK, &td_font,          _countof(td_font)},
                              {_T("TimingDiagram"), _T("FontSize"),         REG_DWORD, REGFLAG_WRITEBACK, &td_fontsize,      1},
                              {_T("TimingDiagram"), _T("FontWeight"),       REG_DWORD, REGFLAG_WRITEBACK, &td_fontweight,    1},
                              {_T("TimingDiagram"), _T("TextLimit"),        REG_DWORD, REGFLAG_WRITEBACK, &td_textlimit,     1} } ;

//                    ,---------------------------------------------------------------------- handle of open key
//                    |        ,------------------------------------------------------------- name of initialization key
//                    |        |           ,------------------------------------------------  name of ini-file
//                    |        |           |                       ,------------------------- pointer to the array of INIVALUEs
//                    |        |           |                       |             ,----------- type of initialization (i.e. registry or ini-file)
//                    |        |           |                       |             |       ,--- number of init values
//                    |        |           |                       |             |       |
static MYINI myini = {MY_HKEY, MY_REG_KEY, MY_INI_SECT _T(".ini"), &inivalue[0], INI_NO, _countof(inivalue)} ;

/*
 Command line parameter definitions
*/
static _TCHAR progname[MAX_PATH+1] ;    /* program name (positional parameter 0) */
static _TCHAR posparm1[PARMSIZE+1] ;    /* positional parameter 1 */
static _TCHAR posparm2[PARMSIZE+1] ;    /* positional parameter 2 */
static _TCHAR strngp[PARMSIZE+1] ;      /* string parameter strng */

/* These are indexes into the CLO table below. Keep them in sync */
typedef enum {PARM_SYSRC=0,
              PARM_SYSRD,
              PARM_SYSRR,
              PARM_SYSRW,
              PARM_SYSIF,
              PARM_SYSID,
              PARM_SYSIR,
              PARM_SYSIW,
              PARM_SYSSX,
              PARM_SYSSY,
              PARM_SYSPP,
              PARM_SYSPT,
              PARM_PROGNAME,
              PARM_LAST} parm_enum ;

//                     ,--------------------------------------------------------------------------- name of parameter without leading "-" or "/". System parameters start with SYSP character.
//                     |              ,------------------------------------------------------------ type of this parameter
//                     |              |            ,----------------------------------------------- pointer to destination for a string parameter
//                     |              |            |             ,--------------------------------- value of this parameter or maximum stringlength in case of a string
//                     |              |            |             |                       ,--------- parameter was found
//                     |              |            |             |                       |      ,-- help text for this parameter
//                     |              |            |             |                       |      |
CLO clo[] =          {{SYSP _T("rc"), PARM_SIMPLE, NULL,         0,                      false, _T("clear registry entries for this program")},
                      {SYSP _T("rd"), PARM_SIMPLE, NULL,         0,                      false, _T("disable use of the registry for this program\r\r\n")
                                                                                                _T("     (will use .ini file instead unless -id specified")},
                      {SYSP _T("rr"), PARM_SIMPLE, NULL,         0,                      false, _T("read initialization values from registry")},
                      {SYSP _T("rw"), PARM_SIMPLE, NULL,         0,                      false, _T("write initialization values back to registry")},
                      {SYSP _T("if"), PARM_STRING, ini_filename, _countof(ini_filename), false, _T("specify ini filename")},
                      {SYSP _T("id"), PARM_SIMPLE, NULL,         0,                      false, _T("disable use of ini file for this program")},
                      {SYSP _T("ir"), PARM_SIMPLE, NULL,         0,                      false, _T("read initialization values from ini file")},
                      {SYSP _T("iw"), PARM_SIMPLE, NULL,         0,                      false, _T("write initialization values back to ini file\r\r\n")
                                                                                                _T("     NOTE: avoid using registry and ini-file at the same time\r\r\n")
                                                                                                _T("     Exception: -rr -iw to export registry to ini-file and\r\r\n")
                                                                                                _T("                -ir -rw to export ini-file to registry")},
                      {SYSP _T("sx"), PARM_INT,    NULL,         0,                      false, _T("set horizontal snap range for dialog")},
                      {SYSP _T("sy"), PARM_INT,    NULL,         0,                      false, _T("set vertical snap range for dialog")},
                      {SYSP _T("pp"), PARM_INT,    NULL,         0,                      false, _T("set process priority for dialog\r\r\n")
                                                                                                _T("       (0:idle, 1:normal, 2:high, 3:realtime)")},
                      {SYSP _T("pt"), PARM_INT,    NULL,         0,                      false, _T("set priority for processing thread\r\r\n")
                                                                                                _T("       (0:idle, 1:lowest, 2:below normal, 3:normal,\r\r\n")
                                                                                                _T("        4:above normal, 5:highest, 6:time critical)")},
                      {NULL,          PARM_POS,    &progname,    _countof(progname),     false, NULL} } ;

size_t clo_size = PARM_LAST ;

/*
 Dialog coordinates definitions
*/
//                       ,----------------------------------------------------------------------------------- ID
//                       |               ,------------------------------------------------------------------- pointer to dialog box handle
//                       |               |                        ,------------------------------------------ additional horizontal offset (required for tabs)
//                       |               |                        |  ,--------------------------------------- additional vertical offset (required for tabs)
//                       |               |                        |  |  ,------------------------------------ left position of control
//                       |               |                        |  |  |  ,--------------------------------- top position of control
//                       |               |                        |  |  |  |  ,------------------------------ right position of control
//                       |               |                        |  |  |  |  |  ,--------------------------- bottom position of control
//                       |               |                        |  |  |  |  |  |     ,--------------------- move speed in x direction (0..1000)
//                       |               |                        |  |  |  |  |  |     |     ,--------------- move speed in y direction (0..1000)
//                       |               |                        |  |  |  |  |  |     |     |     ,--------- resize speed in x direction (0..1000)
//                       |               |                        |  |  |  |  |  |     |     |     |     ,--- resize speed in y direction (0..1000)
//                       |               |                        |  |  |  |  |  |     |     |     |     |
COORDS coordinates[] = {{0,              &MainWindow,             0, 0, 0, 0, 0, 0,    0,    0,    0,    0},  /* first entry is reserved for main dialog */
                        {DLG_STATUS,     &MainWindow,             0, 0, 0, 0, 0, 0,    0, 1000, 1000,    0},
                        {IDEC_INPUT,     &MainWindow,             0, 0, 0, 0, 0, 0,    0,    0, 1000,  500},
                        {IDEC_OUTPUT,    &MainWindow,             0, 0, 0, 0, 0, 0,    0,  500, 1000,  500},
                        {IDTX_OUTPUT,    &MainWindow,             0, 0, 0, 0, 0, 0,    0,  500,    0,    0},
                        {BUT_INFO,       &MainWindow,             0, 0, 0, 0, 0, 0,    0, 1000,    0,    0},
                        {BUT_START,      &MainWindow,             0, 0, 0, 0, 0, 0, 1000, 1000,    0,    0},
                        {BUT_EXIT,       &MainWindow,             0, 0, 0, 0, 0, 0, 1000, 1000,    0,    0}
                       } ;

COORDS *sorted_coord[_countof(coordinates)-1] ;
size_t coord_size = _countof(coordinates) ;
static bool validreg = false ;          /* valid data from registry retrieved */
static bool abort_thread = false ;      /* set true to abort the thread */
static bool thread_suspended = false ;
_TCHAR purpose[] = PURPOSE ;
PTCHAR errorparm = NULL ;
static int32_t comparesize ;            /* number of characters to compare, -1 if all */

/* signal level */
#define LOW 0
#define HIGH 1
#define HIGHIMP 2
#define UNKNOWN 3
#define HIGHORLOW 4
#define RISING 5
#define FALLING 6

/* previous signal position (index in table below) */
#define POSLOW 0
#define POSHIGH 7
#define POSMID 14
#define POSHILO 21
#define NOPOSITION 55 /* any number different from previous list */

static int position[] = {POSLOW,POSHIGH,POSMID,POSMID,POSHILO,POSHIGH,POSLOW} ;

/* Character tables to generate diagrams */
//                              ┌────────────────────────────────────────────────────── low (after low)
//                              │┌───────────────────────────────────────────────────── high (after low)
//                              ││┌──────────────────────────────────────────────────── high impedance (after low)
//                              │││┌─────────────────────────────────────────────────── unknown (after low)
//                              ││││┌────────────────────────────────────────────────── high or low (after low)
//                              │││││┌───────────────────────────────────────────────── rising edge (after low)
//                              ││││││┌──────────────────────────────────────────────── falling edge (after low)
//                              │││││││
//                              │││││││       ┌──────────────────────────────────────── low (after high)
//                              │││││││       │┌─────────────────────────────────────── high (after high)
//                              │││││││       ││┌────────────────────────────────────── high impedance (after high)
//                              │││││││       │││┌───────────────────────────────────── unknown (after high)
//                              │││││││       ││││┌──────────────────────────────────── high or low (after high)
//                              │││││││       │││││┌─────────────────────────────────── rising edge (after high)
//                              │││││││       ││││││┌────────────────────────────────── falling edge (after high)
//                              │││││││       │││││││
//                              │││││││       │││││││       ┌────────────────────────── low (after mid)
//                              │││││││       │││││││       │┌───────────────────────── high (after mid)
//                              │││││││       │││││││       ││┌──────────────────────── high impedance (after mid)
//                              │││││││       │││││││       │││┌─────────────────────── unknown (after mid)
//                              │││││││       │││││││       ││││┌────────────────────── high or low (after mid)
//                              │││││││       │││││││       │││││┌───────────────────── rising edge (after mid)
//                              │││││││       │││││││       ││││││┌──────────────────── falling edge (after mid)
//                              │││││││       │││││││       │││││││
//                              │││││││       │││││││       │││││││       ┌──────────── low (after hilo)
//                              │││││││       │││││││       │││││││       │┌─────────── high (after hilo)
//                              │││││││       │││││││       │││││││       ││┌────────── high impedance (after hilo)
//                              │││││││       │││││││       │││││││       │││┌───────── unknown (after hilo)
//                              │││││││       │││││││       │││││││       ││││┌──────── high or low (after hilo)
//                              │││││││       │││││││       │││││││       │││││┌─────── rising edge (after hilo)
//                              │││││││       │││││││       │││││││       ││││││┌────── falling edge (after hilo)
//                              │││││││       │││││││       │││││││       │││││││
static _TCHAR ct_upper[] =  _T(" ┌  ┌┌ ") _T("┐─┐┐┬┬┐") _T(" ┌  ┌┌ ") _T("┐┬┐┐─┐ ") ;
static _TCHAR ct_middle[] = _T(" │┌┌│↑↓") _T("│ └└│↑↓") _T("┐┘─x┤↑↓") _T("││├├ ↑↓") ;
static _TCHAR ct_lower[] =  _T("─┘┘┘┴┘┴") _T("└   └ └") _T("└   └ └") _T("└┘┘┘─ ┘") ;



/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
static INT_PTR CALLBACK DrawTimingDiagram(HWND hwnd, unsigned msg, WPARAM wParam, LPARAM lParam) ;
static INT_PTR CALLBACK DlgHelp(HWND hwnd, unsigned msg, WPARAM wParam, LPARAM lParam) ;
void GenerateTiming(void *parm) ;
void getstrng(PTCHAR src, PTCHAR dest, int maxlen) ;
void OutputTiming(HWND hwnd, PTCHAR signalname, PTCHAR strng) ;
void OutputLabel(HWND hwnd, PTCHAR strng) ;
void SanityCheck(void) ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void SanityCheck(void)                                                                                          |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Check configuration for validity                                                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void SanityCheck(void)
{
   assert(PARM_LAST == _countof(clo)) ; /* check size of parameter table */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       WinMain                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| MAIN program.                                                                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#pragma warning(push)
#pragma warning(disable : 28251)
int WINAPI WinMain(HINSTANCE this_inst, HINSTANCE prev_inst, LPSTR cmdline, int cmdshow)
/* WinMain generates false warning on 64-bit compilation with Visual Studio 2019 */
#pragma warning(pop)
{
INT_PTR static volatile rv ;
bool startup ;
#ifndef NDEBUG
   SanityCheck() ;
#endif
   UNREFERENCED_PARAMETER(prev_inst) ;
   UNREFERENCED_PARAMETER(cmdline) ;
   UNREFERENCED_PARAMETER(cmdshow) ;
   startup = true ;
   WinLibInit() ;
   errorparm = getparms(GetCommandLine()) ;
   if (clo[PARM_SYSRC].found) clear_registry(&myini) ;
   myini.initype = INI_NO ;
   if (!clo[PARM_SYSRD].found && !clo[PARM_SYSRR].found && !clo[PARM_SYSRW].found && !clo[PARM_SYSIR].found && !clo[PARM_SYSIW].found)
   { /* enable registry read/write if none of the previous parameters was found (as default and for compatibility) */
      myini.initype |= (INI_REGREAD + INI_REGWRITE) ;
   }
   if (clo[PARM_SYSRD].found && !clo[PARM_SYSRR].found && !clo[PARM_SYSRW].found && !clo[PARM_SYSIR].found && !clo[PARM_SYSIW].found && !clo[PARM_SYSID].found)
   { /* enable ini file read/write if registry is disabled and none of the previous parameters was found
        use "-rd -id" to use neither registry nor ini file */
      myini.initype |= (INI_FILEREAD + INI_FILEWRITE) ;
   }
   if (clo[PARM_SYSRR].found) myini.initype |= INI_REGREAD ;
   if (clo[PARM_SYSRW].found) myini.initype |= INI_REGWRITE ;
   if (clo[PARM_SYSIR].found) myini.initype |= INI_FILEREAD ;
   if (clo[PARM_SYSIW].found) myini.initype |= INI_FILEWRITE ;
   if (clo[PARM_SYSRD].found) myini.initype &= ~(INI_REGREAD + INI_REGWRITE) ;
   if (clo[PARM_SYSID].found) myini.initype &= ~(INI_FILEREAD + INI_FILEWRITE) ;
   if (clo[PARM_SYSIF].found) myini.inifilename = ini_filename ;
   if (clo[PARM_SYSSX].found) snap_x = clo[PARM_SYSSX].value ;
   if (clo[PARM_SYSSY].found) snap_y = clo[PARM_SYSSY].value ;
   if (startup)
   { /* start windowed version */
      if (clo[PARM_SYSPP].found) set_process_prio(clo[PARM_SYSPP].value) ;
      MyInstance = this_inst ;
      InitCommonControls() ;
      rv = DialogBox(this_inst, MAKEINTRESOURCE(DLG_TIMINGDIAGRAM), 0, (DLGPROC)DrawTimingDiagram) ;
   }
   return(0);
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       DrawTimingDiagram                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Processes messages for the DLG_TIMINGDIAGRAM dialog (see "DialogProc" function)                                       |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
INT_PTR CALLBACK DrawTimingDiagram(HWND hwnd, unsigned msg, WPARAM wParam, LPARAM lParam)
{ /* Latest MS-Doc says that INT_PTR is the correct data type here, not BOOL */
HICON static hIconS ;
HICON static hIconB ;
INT_PTR retcd ; /* return true, if the message is processed here and false, if it is not (except WM_INITDIALOG) */
bool msg_processed ;
RECT rect ;
HFONT DlgFont ;
HANDLE static hThread ;

   retcd = false ; /* return false if we don't process the message */
   msg_processed = false ; /* message was not precessed */
   switch(msg)
   {
   case WM_INITDIALOG:
      retcd = true ; /* we don't set the focus here */
      msg_processed = true ; /* message was processed */
      MainWindow = hwnd ;
      hIconS = (HICON)LoadImage(MyInstance, MAKEINTRESOURCE(IDI_ICON), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
      hIconB = (HICON)LoadImage(MyInstance, MAKEINTRESOURCE(IDI_ICON), IMAGE_ICON, 32, 32, LR_DEFAULTCOLOR);
      if (hIconS != NULL) SendMessage(MainWindow, WM_SETICON, ICON_SMALL, (LPARAM)hIconS);
      if (hIconB != NULL) SendMessage(MainWindow, WM_SETICON, ICON_BIG, (LPARAM)hIconB);

      /* create status bar */
      StatusWindow = CreateWindowEx(0,                                      /* extended window style */
                                    STATUSCLASSNAME,                        /* pointer to registered class name */
                                    NULL,                                   /* pointer to window name */
                                    WS_CHILD | WS_VISIBLE | SBARS_SIZEGRIP, /* window style */
                                    0, 0, 0, 0,                             /* horizontal and vertical position of window, width and height */
                                    MainWindow,                             /* handle to parent or owner window */
                                    (HMENU)DLG_STATUS,                      /* handle to menu, or child-window identifier */
                                    GetModuleHandle(NULL),                  /* handle to application instance */
                                    NULL);                                  /* pointer to window-creation data */

// save original coordinates of dialog
      SaveCoordinates() ;

      validreg = get_registry(&myini) ; // get registry content

      if (validreg)
      {
         MoveWindow(
         MainWindow,            /* handle of window */
         registry_x,            /* horizontal position */
         registry_y,            /* vertical position */
         registry_width,        /* width */
         registry_height,       /* height */
         true);                 /* repaint flag */
      }


      DlgFont = CreateFont(
          td_fontsize,          /* logical height of font */
          0,                    /* logical average character width */
          0,                    /* angle of escapement */
          0,                    /* base-line orientation angle */
          td_fontweight,        /* font weight */
          false,                /* italic attribute flag */
          false,                /* underline attribute flag */
          false,                /* strikeout attribute flag */
          ANSI_CHARSET,         /* character set identifier */
          OUT_DEFAULT_PRECIS,   /* output precision */
          CLIP_DEFAULT_PRECIS,  /* clipping precision */
          DEFAULT_QUALITY,      /* output quality */
          DEFAULT_PITCH,        /* pitch and family */
          td_font               /* pointer to typeface name string */
      ) ;

      SendMessage(GetDlgItem(hwnd, IDEC_INPUT), WM_SETFONT, (WPARAM) DlgFont, MAKELPARAM(false, 0));
      SendMessage(GetDlgItem(hwnd, IDEC_OUTPUT), WM_SETFONT, (WPARAM) DlgFont, MAKELPARAM(false, 0));
      SendMessage(GetDlgItem(hwnd, IDEC_INPUT), EM_SETLIMITTEXT, (WPARAM) td_textlimit, 0);
      SendMessage(GetDlgItem(hwnd, IDEC_OUTPUT), EM_SETLIMITTEXT, (WPARAM) td_textlimit, 0);

      /* display any illegal parameter */
      if (errorparm != NULL)
      {
         LoadString(MyInstance, STR_INVALIDPARM, strng1, _countof(strng1));
         _stprintf_s(strng2, _countof(strng2), strng1, errorparm) ;
         SendMessage(StatusWindow, SB_SETTEXT, STATUSMESSAGE, (LPARAM) strng2) ;
      }
      _stprintf_s(strng2, _countof(strng2), _T("%s %u.%2.2u"), PROGRAMNAME _T(" - ") INTSIZE_T DEBUG_T _T("-Version"), PROGRAMVERSION/100, PROGRAMVERSION%100) ;
      SendMessage(MainWindow, WM_SETTEXT, 0, (LPARAM) strng2);

   break ;

   case WM_COMMAND:
      switch(HIWORD(wParam))
      {
      case BN_CLICKED:
         switch(LOWORD(wParam))
         {
            case BUT_INFO:
               retcd = true ;
               msg_processed = true ; /* message was processed */
               _stprintf_s(strng1, _countof(strng1), _T("%s - Version %u.%2.2u"), PROGRAMNAME, PROGRAMVERSION/100, PROGRAMVERSION%100) ;
               PrintProgramInfo(DLG_HELPLIST, strng1, _T(__DATE__), _T(__TIME__)) ;
            break ;

            case IDOK:
            case BUT_START:
               retcd = true ;
               msg_processed = true ; /* message was processed */
               hThread = (HANDLE) _beginthread(GenerateTiming, 0, NULL) ;
            break ;

            case IDABORT:
            case IDCANCEL:
            case BUT_EXIT:
               retcd = true ;
               msg_processed = true ; /* message was processed */
               GetWindowRect(MainWindow, &rect) ; /* get current dialog coordinates */
               registry_x = rect.left ;
               registry_y = rect.top ;
               registry_width = rect.right - rect.left ;
               registry_height = rect.bottom - rect.top ;
               if (registry_x < 0) registry_x = 0 ;
               if (registry_y < 0) registry_y = 0 ;
               registry_version = PROGRAMVERSION ;    /* version of program which wrote the registry */
               put_registry(&myini) ;                 /* store the dialog properties in registry */
               EndDialog(MainWindow, true) ;          /* exits the dialog box */
            break ;
         }
      break ;
      }
   break ;

   }

   if (!msg_processed)
   { /* If message was not processed here, try WinLib. */
      retcd = WinLibMessage(msg, wParam, lParam) ;
   }

   return(retcd) ;
} /* DrawTimingDiagram */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void GenerateTiming(void *parm)                                                                                 |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read input from input control and convert it to a timing diagram in the output control.                               |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void GenerateTiming(void *parm)
{
int i, j ;
INT_PTR numlines ;
int linelength ;
static _TCHAR signal_name[MAX_SIGNALNAME] ;
HWND hwnd ;

   UNREFERENCED_PARAMETER(parm) ;
   hwnd = MainWindow ;
   SendMessage(GetDlgItem(hwnd, IDEC_OUTPUT), EM_SETSEL, 0, -1); /* clear output window */
   SendMessage(GetDlgItem(hwnd, IDEC_OUTPUT), EM_REPLACESEL, (bool) false, (WPARAM) _T("")) ;

   numlines = SendMessage(GetDlgItem(hwnd, IDEC_INPUT), EM_GETLINECOUNT, 0, 0) ;
   signal_name[0] = _T('\0') ;
   for (i=0; i<numlines;i++)
   {
      linelength = Edit_GetLine(GetDlgItem(hwnd, IDEC_INPUT), i, strng1, _countof(strng1)) ; /* returns number of characters, no terminating 0 */
      strng1[linelength] = _T('\0') ;
      for (j=0; (j<linelength) && (strng1[j] == _T(' ')); j++) ;
      if (strng1[j] == COMMANDCHAR)
      { /* command found, ignore anything else */
         switch (strng1[j+1])
         {
         case _T('s'): /* signal name follows */
         case _T('S'):
            getstrng(&strng1[j+2], signal_name, _countof(signal_name)) ;
         break ;

         case _T('t'): /* timing string follows */
         case _T('T'):
            getstrng(&strng1[j+2], strng2, _countof(strng2)) ;
            OutputTiming(hwnd, signal_name, strng2) ;
         break ;

         case _T('l'): /* label string follows */
         case _T('L'):
            getstrng(&strng1[j+2], strng2, _countof(strng2)) ;
            OutputLabel(hwnd, strng2) ;
         break ;

         }
      }
   }
}

/*
l: low
h: high
z: high impedance
x: unknown
X: high or low
r: rising edge
f: falling edge
 : space
~: break

sample input:

$s "clk0"
$t "         lllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhl ~ hhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhh"
$l "            |tclk |"
$s "srd"
$t "         xxxxxXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX ~ XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXxxxxxxxxxxxxxxxxxx"
$l "                  ^ sampled here"
$l "              |---| tsu                                                ~                                          |--| th"
$s "saddr"
$t "         xxxxxXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX ~ XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXxxxxxxxxxxxxxxxxxx"
$l "                  |-------| > 1 * tclk"
$s "sena"
$t "         lllllllllllllllllhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh ~ hhhhhhhhhhhhhhhhhhhhhhhhhhhlllllllllllllllllllllllllllllllllll"
$l "                              |-----| second rising clk0 edge after sena = 1"
$s "rising_sena (internal)"
$t "         lllllllllllllllllllllllllllrhhhhhllllllllllllllllllllllllllll ~ llllllllllllllllllllllllllllllllllllllllllllllllllllllllllllll"
$l "                                          ^ transfer read register data to serial output buffer"
$s "falling_sena (internal)"
$t "         lllllllllllllllllllllllllllllllllllllllllllllllllllllllllllll ~ lllllllllllllllllllllllllllllllllllrhhhhhlllllllllllllllllllll"
$l "                                                                    transfer serial input buffer data to register ^"
$s "clk0"
$t "         lllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhl ~ hhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhhlllhhh"
$l "                                          |-| > 0"
$l "                                                |---------| > 1 * tclk"
$s "sclk"
$t "         lllllllllllllllllllllllllllllllllllhhhhhhhhhhhhhhllllllllllll ~ hhhhhhhhhhhhhhhhhlllllllllllllllllllllllllllllllllllllllllllll"
$s "rising_sclk (internal)"
$t "         lllllllllllllllllllllllllllllllllllllllllllllrhhhhhllllllllll ~ lllllllllllrhhhhhlllllllllllllllllllllllllllllllllllllllllllll"
$l "                                                            ^ shift next bit in/out"
$l "                                                        |tsu| th|"
$s "sdi"
$t "         xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxXXXXXXXXxxxxxx ~ xxxxxxxxxxxxxXXXXXXXXxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
$s "sdo"
$t "         zzzzzzzzzzzzzzzzzXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXxxXXXXXXXX ~ XXXXXXXXXXXXXXXXXxxXXXXXXXXzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz"
$l ""
$l "Note: tsu and th depend on the device (see simulation). 10 ns (min) should be enough for all."


sample output:
            ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐ ~ ──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──
clk0        │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │ ~   │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │
         ───┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └ ~   └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘
            |tclk |
              ┌─────────────────────────────────────────────────────── ~ ────────────────────────────────────────────┐
srd      xxxxx┤                                                        ~                                             ├xxxxxxxxxxxxxxxxx
              └─────────────────────────────────────────────────────── ~ ────────────────────────────────────────────┘
                  ↑ sampled here
              |---| tsu                                                ~                                          |--| th
              ┌─────────────────────────────────────────────────────── ~ ────────────────────────────────────────────┐
saddr    xxxxx┤                                                        ~                                             ├xxxxxxxxxxxxxxxxx
              └─────────────────────────────────────────────────────── ~ ────────────────────────────────────────────┘
                  |-------| > 1 * tclk
                          ┌─────────────────────────────────────────── ~ ───────────────────────────┐
sena                      │                                            ~                            │
         ─────────────────┘                                            ~                            └──────────────────────────────────
                              |-----| second rising clk0 edge after sena = 1
                                    ┌─────┐                            ~
rising_sena (internal)              ↑     │                            ~
         ───────────────────────────┘     └─────────────────────────── ~ ──────────────────────────────────────────────────────────────
                                          ↑ transfer read register data to serial output buffer
                                                                       ~                                    ┌─────┐
falling_sena (internal)                                                ~                                    ↑     │
         ───────────────────────────────────────────────────────────── ~ ───────────────────────────────────┘     └────────────────────
                                                                    transfer serial input buffer data to register ↑
            ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐ ~ ──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──┐  ┌──
clk0        │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │ ~   │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │  │
         ───┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └ ~   └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘  └──┘
                                          |-| > 0
                                                |---------| > 1 * tclk
                                            ┌─────────────┐            ~ ─────────────────┐
sclk                                        │             │            ~                  │
         ───────────────────────────────────┘             └─────────── ~                  └────────────────────────────────────────────
                                                      ┌─────┐          ~            ┌─────┐
rising_sclk (internal)                                ↑     │          ~            ↑     │
         ─────────────────────────────────────────────┘     └───────── ~ ───────────┘     └────────────────────────────────────────────
                                                            ↑ shift next bit in/out
                                                        |tsu| th|
                                                        ┌───────┐      ~              ┌───────┐
sdi      xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx┤       ├xxxxx ~ xxxxxxxxxxxxx┤       ├xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
                                                        └───────┘      ~              └───────┘
                          ┌─────────────────────────────────┐ ┌─────── ~ ─────────────────┐ ┌───────┐
sdo      ─────────────────┤                                 ├x┤        ~                  ├x┤       ├──────────────────────────────────
                          └─────────────────────────────────┘ └─────── ~ ─────────────────┘ └───────┘

Note: tsu and th depend on the device (see simulation). 10 ns (min) should be enough for all.

*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void getstrng(PTCHAR src, PTCHAR dest, int maxlen)                                                              |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Get formatted string from src and copy to dest. Remove any " or '                                                     |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void getstrng(PTCHAR src, PTCHAR dest, int maxlen)
{
int i, j ;
   i = 0 ;
   j = 0 ;
   while ((src[i] != _T('\"')) && (src[i] != _T('\0'))) i++ ; /* find leading " */
   if (src[i] != _T('\0'))
   {
      i++ ;
      while ((src[i] != _T('\"')) && (src[i] != _T('\0')))
      {
         if (j < maxlen) dest[j] = src[i] ;
         j++ ; /* next dest */
         i++ ; /* next src */
      }

   }
   dest[j] = _T('\0') ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void OutputLabel(hwnd, strng2)                                                                                  |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Generate a label                                                                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void OutputLabel(HWND hwnd, PTCHAR strng)
{
int strpos ;
static _TCHAR local_strng[STRINGSIZE] ;
int strngindex ;

   strpos = 1 ;
   strngindex = 0 ;
   while (strng[strpos-1] != _T('\0'))
   {
      switch (strng[strpos-1])
      {
      case _T('^'):
         local_strng[strngindex++] = _T('↑') ;
      break ;

      default:
         local_strng[strngindex++] = strng[strpos-1] ;
      break ;
      }

      strpos++ ;
   }
   local_strng[strngindex++] = _T('\r') ;
   local_strng[strngindex++] = _T('\n') ;
   local_strng[strngindex++] = _T('\0') ;
   SendMessage(GetDlgItem(hwnd, IDEC_OUTPUT), EM_REPLACESEL, (bool) false, (WPARAM) local_strng) ;

}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void OutputTiming(HWND hwnd, PTCHAR signalname, PTCHAR strng)                                                   |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Generate a signal                                                                                                     |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void OutputTiming(HWND hwnd, PTCHAR signalname, PTCHAR strng)
{
size_t prev ;
PTCHAR ct ;
size_t line ;
size_t namelen ;
size_t strpos ;
static _TCHAR local_strng[STRINGSIZE] ;
size_t strngindex ;

   for (line=0; line<3; line++)
   {
      ct = ct_upper ;
      namelen = 0 ;
      strngindex = 0 ;
      if (line == 1)
      {
         ct = ct_middle ;
         namelen = _tcsclen(signalname) ;
         _tcscpy_s(local_strng, _countof(local_strng), signalname) ;
         strngindex += namelen ;
      }
      if (line == 2) ct = ct_lower ;
      strpos = 1 ;
      prev = NOPOSITION ;
      while (strng[strpos-1] != _T('\0'))
      {
         if (strpos > namelen)
         {
            switch (strng[strpos-1])
            {
            case _T('l'):
               if (prev == NOPOSITION) prev = position[LOW] ;
               local_strng[strngindex++] = ct[LOW+prev] ;
               prev = position[LOW] ;
            break ;

            case _T('h'):
               if (prev == NOPOSITION) prev = position[HIGH] ;
               local_strng[strngindex++] = ct[HIGH+prev] ;
               prev = position[HIGH] ;
            break ;

            case _T('z'):
               if (prev == NOPOSITION) prev = position[HIGHIMP] ;
               local_strng[strngindex++] = ct[HIGHIMP+prev] ;
               prev = position[HIGHIMP] ;
            break ;

            case _T('x'):
               if (prev == NOPOSITION) prev = position[UNKNOWN] ;
               local_strng[strngindex++] = ct[UNKNOWN+prev] ;
               prev = position[UNKNOWN] ;
            break ;

            case _T('X'):
               if (prev == NOPOSITION) prev = position[HIGHORLOW] ;
               local_strng[strngindex++] = ct[HIGHORLOW+prev] ;
               prev = position[HIGHORLOW] ;
            break ;

            case _T('r'):
               if (prev == NOPOSITION) prev = position[LOW] ;
               local_strng[strngindex++] = ct[RISING+prev] ;
               prev = position[RISING] ;
            break ;

            case _T('f'):
               if (prev == NOPOSITION) prev = position[HIGH] ;
               local_strng[strngindex++] = ct[FALLING+prev] ;
               prev = position[FALLING] ;
            break ;

            case _T(' '):
               local_strng[strngindex++] = _T(' ') ;
            break ;

            case _T('~'):
               local_strng[strngindex++] = _T('~') ;
               prev = NOPOSITION ;
            break ;
            }
         }

         strpos++ ;
      }
      local_strng[strngindex++] = _T('\r') ;
      local_strng[strngindex++] = _T('\n') ;
      local_strng[strngindex++] = _T('\0') ;
//      SendMessage(GetDlgItem(hwnd, IDEC_OUTPUT), EM_SETSEL, 0, -1); /* clear output window */
      SendMessage(GetDlgItem(hwnd, IDEC_OUTPUT), EM_REPLACESEL, (bool) false, (WPARAM) local_strng) ;

   }
}

